# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.12.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

import sympy as sp
sp.init_printing()



# ## Coordinates and metric

coords = sp.symbols(r't r \theta \phi', real=True)
t, r, θ, ϕ = coords
N = len(coords)
A = sp.Function('A')(r)
B = sp.Function('B')(r)

g = sp.diag(-A**2, B**2, B**2 * r**2, B**2 * r**2 * sp.sin(θ)**2)
ginv = g.inv()
display(g)
display(ginv)


# +
#Φ = sp.Function(r'\Phi', real=True)(r)
#Λ = sp.Function(r'\Lambda', real=True)(r)
#g = sp.diag(-sp.exp(2*Φ), sp.exp(2*Λ), r**2, r**2 * sp.sin(θ)**2)
#ginv = g.inv()
#display(g)
#display(ginv)

# +
#Φ = sp.Function(r'\Phi', real=True)(r)
#m = sp.Function(r'm', real=True)(r)
#g = sp.diag(-sp.exp(2*Φ), 1/(1-2*m/r), r**2, r**2 * sp.sin(θ)**2)
#ginv = g.inv()
#display(g)
#display(ginv)
# -

# ## Christoffels

# +
def christoffels(coords, g, ginv, simplify=False):
    N = len(coords)
    Γs = [ sp.zeros(N,N) for _ in range(N) ]
    oh = sp.Rational(1,2)
    for (i, _) in enumerate(coords):
        Γ = Γs[i]
        for k in range(N):
            for l in range(N):
                for m in range(N):
                    Γ[k,l] += oh * ginv[i,m] * \
                        (sp.diff(g[m,k],coords[l]) + sp.diff(g[m,l],coords[k]) - sp.diff(g[k,l],coords[m]))
        if simplify:
            Γ = Γ.simplify()
    return Γs
    
Γ4s = christoffels(coords, g, ginv, simplify=True)
display(Γ4s)
# -

# ## Induced metric

from sympy.physics.quantum.tensorproduct import TensorProduct

n = sp.Matrix([-A, 0, 0, 0])
display(n)

γ4 = g + TensorProduct(n.transpose(),n)
display(γ4)

γ_coords = coords[1:]
display(γ_coords)
γ = γ4[1:,1:]
γinv = γ.inv()
display(γ)
display(γinv)

γ4ud = ginv * γ4
display(γ4ud)

Γs = christoffels(γ_coords, γ, γinv)
display(Γs)

η_coords = coords[1:]
display(η_coords)
η = sp.diag(1, r**2, r**2 * sp.sin(θ)**2)
ηinv = η.inv()
display(η)
display(ηinv)
display(η.det())
η_Γs = christoffels(η_coords, η, ηinv)
display(η_Γs)

# $$
#     \gamma_{ij} = r^{-2/3} \eta_{ij}
# $$

γhat_coords = coords[1:]
display(γhat_coords)
γhat = r**(sp.Rational(-2,3)) * η
γhatinv = γhat.inv()
display(γhat)
display(γhatinv)
display(γhat.det())
γhat_Γs = christoffels(γhat_coords, γhat, γhatinv)
display(γhat_Γs)


# ## Extrinsic curvature
#
# $$
#     K_{ij} = - {^{(4)}\gamma}_{i}^{k} {^{(4)}\gamma}_{j}^{l} \nabla_{k} n_{l}
# $$

def extrinsic_curvature(coords, γ4ud, Γ4s, n):
    N = len(coords)
    K = sp.zeros(N,N)
    
    for i in range(N):
        for j in range(N):
            K[i,j] = 0
            for k in range(N):
                for l in range(N):
                    γik = γ4ud[i,k]; γjl = γ4ud[j,l]
                    K[i,j] += sp.diff(n[l], coords[k])
                    for m in range(N):
                        Γ = Γ4s[m]
                        K[i,j] -= Γ[k,l] * n[m]
                    K[i,j] *= γik * γjl
            K[i,j] *= -1
    
    return K
K = extrinsic_curvature(coords, γ4ud, Γ4s, n)
display(γ4ud)
display(n)
display(K)

sphere_coords = coords[1:]
sphere_g = sp.diag(1, r**2, r**2 * sp.sin(θ)**2)
sphere_ginv = sphere_g.inv()
display(sphere_g)
display(sphere_ginv)

sphere_Γs = christoffels(sphere_coords, sphere_g, sphere_ginv)
display(sphere_Γs)

from IPython.display import Math
N = len(sphere_coords)
for (c, Γ) in zip(sphere_coords, sphere_Γs):
    for i in range(N):
        for j in range(N):
            Γij = Γ[i,j]
            if Γij:
                ci = sphere_coords[i];
                cj = sphere_coords[j];
                display(Math(f"\Gamma^{{ {c} }}_{{ {ci} {cj} }} = " + sp.latex(Γij)))
                # print(sp.latex(Γij))


# ## Metric compatibility
# $$
#     \nabla_\lambda g_{\mu\nu} = 0
# $$

def test_metric_compatibility(coords, g, Γs, simplify=False):
    N = len(coords)
    diffgs = [ sp.zeros(N,N) for _ in range(N) ]
    for (i, diffg) in enumerate(diffgs):
        for j in range(N):
            for k in range(N):
                diffg[j,k] = sp.diff(g[j,k],coords[i])
                for (l, Γl) in enumerate(Γs):
                    diffg[j,k] -= g[j,l] * Γl[i,k] + g[k,l] * Γl[i,j]
            if simplify:
                diffg = diffg.simplify()
    return diffgs
sphere_diffg = test_metric_compatibility(sphere_coords, sphere_g, sphere_Γs)
display(sphere_diffg)


