module Sgibbs

using dg1d
using Random
using Plots
using LinearAlgebra


function s_map(x, ξ, d, k)
  s = x
  for (i, ξi) in enumerate(ξ)
    if (x > ξi)
      di = d[i]
      # s += k * (di + 1.0)
      s += k * (di + 1.0)
    end
  end
  return s
end

function difference_test_sawtooth(n, k)

  x,w = dg1d.LGL.rule(n)

  d = [3.0]; ξ = [0.0]
  s(x) = s_map(x, ξ, d, k)
  f(x) = (x > ξ[1]) ? x - 1.0 : x + 1.0

  sx = s.(x)
  fx = f.(x)

  D = dg1d.first_derivative_matrix(sx)
  dg1d.purge_zeros!(D)
  dfx = D * fx
  dfx_ana = ones(size(x))

  diff = @. dfx - dfx_ana
  err = maximum(diff[:])

  return err
end


function random_pos()
  RNG = MersenneTwister()
  return 0.1 * (rand(RNG, 1) * 2.0 .- 1.0)
  # return rand(1) * 2.0 .- 1.0
end


function difference_test_kink(n, k)

  x,w = dg1d.LGL.rule(n)

  # ξ = [0.0]; d = [0.0]
  ξ = [random_pos()...]; d = [0.0]
  s(x) = s_map(x, ξ, d, k)
  f(x) = (x > ξ[1]) ? (x-ξ[1]) * 1.0 : 0.0

  sx = s.(x)
  fx = f.(x)

  D = dg1d.first_derivative_matrix(sx)
  dg1d.purge_zeros!(D)
  dfx = D * fx
  dfx_ana = ones(size(x))

  diff = @. dfx - dfx_ana
  err = maximum(diff[:])

  # println("err = $err")
  # println("pos = $(ξ...)")
  # println("min dx = $(minimum(abs.(x .- ξ[1])))")

  return err
end



function difference_test_analytical(n, k)

  x,w = dg1d.LGL.rule(n)

  f(x) = exp(x) - x^5 + 3.0 * x^2
  df(x) = exp(x) - 5.0 * x^4 + 6.0 * x

  fx = f.(x)
  dfx_ana = df.(x)

  D = dg1d.first_derivative_matrix(x)
  dg1d.purge_zeros!(D)
  dfx = D * fx

  diff = @. dfx - dfx_ana
  err = maximum(diff[:])

  return err
end


# function max_error_test(f, n, k)
#
#   x,w = dg1d.LGL.rule(n)
#
#   d = [2.0]; ξ = [0.0]
#   s(x) = s_map(x, ξ, d, k)
#   f(x) = (x > ξ[1]) ? x - 1.0 : x + 1.0
#
#   sx = s.(x)
#   fx = f.(x)
#
#   D = dg1d.first_derivative_matrix(sx)
#   dg1d.purge_zeros!(D)
#   dfx = D * fx
#   dfx_ana = ones(size(x))
#
#   diff = @. dfx - dfx_ana
#   err = maximum(diff[:])
#
#   return err
# end
#
#
# function kink(x, ξ)
#
#   f = 0.0
#   df = 0.0
#   if (x >= ξ)
#     f += (x-ξ) * 5
#     df = 5
#   end
#
#   return f
# end


function difference_test_bump(n, k)

  ξ = [ 0.4, 0.5 ]

  function f(x)
    result = 0.0
    if x >= ξ[1] && x <= ξ[2]
      slope = 0.0
      result = slope * (x - ξ[1]) + 1.0
    end
    return result
  end

  function df(x)
    result = 0.0
    if x >= ξ[1] && x <= ξ[2]
      slope = 0.0
      result = slope
    end
    return result
  end

  d = f.(ξ)
  x,w = dg1d.LGL.rule(n)
  s(x) = s_map(x, ξ, d, k)

  sx = s.(x)
  fx = f.(x)
  D = dg1d.first_derivative_matrix(sx)
  dg1d.purge_zeros!(D)
  dfx = D * fx

  dfx_ana = df.(x)
  diff = @. dfx - dfx_ana
  err = maximum(diff[:])

  return err
end


function difference_test(which_test::Symbol, rng_N, rng_K)

  diffs = zeros(length(rng_N), length(rng_K))
  for (i,n) in enumerate(rng_N), (j,k) in enumerate(rng_K)
    if which_test == :sawtooth
      diffs[i,j] = difference_test_sawtooth(n,k)
    elseif which_test == :bump
      diffs[i,j] = difference_test_bump(n,k)
    elseif which_test == :kink
      diffs[i,j] = difference_test_kink(n,k)
    elseif which_test == :analytical
      diffs[i,j] = difference_test_analytical(n,k)
    else
      error("Invalid test requested!")
    end
  end

  return diffs
end


function visz_difference_test()

  which_test = :kink
  rng_N = collect(5:5:45)
  rng_K = collect(0:10:80)

  diffs = difference_test(which_test, rng_N, rng_K)
  diffs[@.(diffs <= 0.0)] .= 1e-16

  pltcfg = "set logscale xy; set grid"

  Plots.gr()
  plt_N = Plots.plot(xlabel="K", ylabel="log(max|error|)")
  for (i,N) in enumerate(rng_N)
    plot!(plt_N, rng_K, abs.(diffs[i,:]), yaxis=(:log), label="N = $N")
  end

  plt_K = Plots.plot(xlabel="N", ylabel="log(max|error|)")
  for (i,K) in enumerate(rng_K)
    plot!(plt_K, rng_K, diffs[:,i], yaxis=(:log), label="K = $K")
  end

  Plots.plot(plt_N, plt_K, layout=(2,1))

end


function test_sawtooth(n, k)

  x,w = dg1d.LGL.rule(n)

  d = [2.0]; ξ = [0.0]
  s(x) = s_map(x, ξ, d, k)
  f(x) = (x > ξ[1]) ? x - 1.0 : x + 1.0

  sx = s.(x)
  fx = f.(x)

  xx = collect(range(-1.0, 1.0, length=200))
  fxx = f.(xx)

  D = dg1d.first_derivative_matrix(sx)
  dg1d.purge_zeros!(D)
  dfx = D * fx
  dfxx = ones(size(xx))

  Plots.gr()
  plt = Plots.plot()
  plot!(plt, xx, fxx)
  plot!(plt, x, fx)
  plot!(plt, xx, dfxx)
  plot!(plt, x, dfx)
  display(plt)

end

function test_bump(n, k)

  x,w = dg1d.LGL.rule(n)
  a = -0.75; b = 0.5

  function f(x)
    result = 0.0
    if (x < a)
      result = - x^4 + 3.0 * x
    elseif (x >= a && x <= b)
      # result = sin(x) + x^2
      result = x^2 + exp(x)
    else
      result = cos(2.0*π/0.5*x)
    end
    return result
  end

  function df(x)
    if (x < a)
      result = - 4.0 * x^3 + 3.0
    elseif (x >= a && x <= b)
      # result = cos(x) + 2.0 * x
      result = 2.0 * x + exp(x)
    else
      result = - 2.0 * π / 0.5 * sin(2.0*π/0.5*x)
    end
    return result
  end

  d = [ f(a), f(b) ]
  ξ = [ a, b ]

  s(x) = s_map(x, ξ, d, k)

  sx = s.(x)
  fx = f.(x)

  xx = collect(range(-1.0, 1.0, length=1000))
  fx_ana = f.(xx)
  dfx_ana = df.(xx)

  D = dg1d.first_derivative_matrix(x)
  dg1d.purge_zeros!(D)
  dfx = D * fx
  E = dg1d.first_derivative_matrix(sx)
  dg1d.purge_zeros!(E)
  dfx_impr = E * fx

  Plots.gr()
  plt = Plots.plot()
  plot!(plt, xx, fx_ana)
  plot!(plt, x, fx, marker=(:circle, 2), markerstrokewidth=0)
  plot!(plt, xx, dfx_ana)
  plot!(plt, x, dfx_impr, marker=(:star6, 2), markerstrokewidth=0)
  display(plt)

end



function test_kink(n, k, r)

  x,w = dg1d.LGL.rule(n)
  max_dx0 = maximum(diff(x))
  r = 0.5 * max_dx0
  x0 = 2 * r * rand() - r
  display(max_dx0)
  # x0 = 2 * rand() - 1

  a = 1e0

  brec = dg1d.BernsteinReconstruction(x)

  function f(x)
    result = 0.0
    if (x < x0)
      result = - sin(4*π*x) + 1
    else
      result = log(x+1) + 1
      # result = a * (x-x0) + 1
    end
    return result
  end

  function df(x)
    if (x < x0)
      result = - 4*π*cos(4*π*x)
    else
      result = 1/(x+1)
      # result = a
    end
    return result
  end

  d = [ f(0.0) ]
  ξ = [ 0.0 ]

  s(x) = s_map(x, ξ, d, k)

  sx = s.(x)
  fx = f.(x)

  xx = collect(range(-1.0, 1.0, length=1000))
  fx_ana = f.(xx)
  dfx_ana = df.(xx)

  D = dg1d.first_derivative_matrix(x)
  dg1d.purge_zeros!(D)
  dfx = D * fx
  E = dg1d.first_derivative_matrix(sx)
  dg1d.purge_zeros!(E)
  bfx = deepcopy(fx)
  # bw = 0.0
  # dg1d.reconstruct!(bfx, bw, brec)
  bw = 1.0
  B = (1 - bw) * brec.V * brec.P + bw * diagm(ones(n))
  # invB = inv(B)
  # bfx = bw * fx + (1.0 - bw) * bfx
  # dfx_impr = E * B * bfx
  dfx_impr = E * bfx

  Plots.gr()
  plt = Plots.plot()
  plot!(plt, xx, fx_ana, label="analytic")
  plot!(plt, x, bfx, marker=(:circle, 2), markerstrokewidth=0, label="numeric")
  plot!(plt, xx, dfx_ana, label="derivative analytic")
  plot!(plt, x, dfx_impr, marker=(:star6, 2), markerstrokewidth=0, label="derivative numeric")
  display(plt)

end



function visz_nodes(k)
  
  Npts_max = 20

  cfg = "set yrange [-0.1:1.1]; set xrange [-1.1:1.1]; set grid; unset key"
  dy = 1.0 / (Npts_max-2)

  # s = :default
  # dg1d.reset_session(s)
  # dg1d.start_session(s)
  # dg1d.cmd_session(cfg, s)
  # for (i,Npts) in enumerate(2:Npts_max)
  #   z, _  = dg1d.LGL.rule(Npts)
  #   fz    = (i-1) * dy * ones(size(z))
  #   dg1d.plot(hcat(z, fz), plotopts="w p ps 2 pt 7")
  # end
  # dg1d.end_session(s)
  
  # d = [2.0, 1.0]; ξ = [-0.25, 0.6]
  d = [2.0]; ξ = [-0.25]
  map(x) = s_map(x, ξ, d, k)
  xmax = map(1.0)
  cfg = "set yrange [-0.1:1.1]; set xrange [-1.1:$xmax]; set grid; unset key"

  Plots.gr()
  plt = plot()
  for (i,Npts) in enumerate(2:Npts_max)
    z, _  = dg1d.LGL.rule(Npts)
    sz = map.(z)
    fsz    = (i-1) * dy * ones(size(sz))
    scatter!(plt, sz, fsz, marker=(:circle,2), markerstrokewidth=0)
  end
  display(plt)

  return
end



function visz_lagrange(k)

  Npts = 5
  z, _  = dg1d.LGL.rule(Npts)
  xx = collect(range(-1.0, 1.0, length=200))

  cfg = "set grid"
  arrows = ""
  for zi in z
    arrows *= "set arrow from $zi,-100.0 to $zi,100.0 nohead back;"
  end

  s = :normal
  dg1d.reset_session(s)
  dg1d.start_session(s)
  dg1d.cmd_session(cfg, s)
  dg1d.cmd_session(arrows, s)

  for i = 1:Npts
    lxx = map(x->dg1d.lagrange_poly(i, x, z), xx)
    dg1d.plot(hcat(xx, lxx), plotopts="w l title 'i = $i'", session=s)
  end

  dg1d.end_session(s)
  
  d = [2.0, 1.0]; ξ = [-0.25, 0.6]
  S(x) = s_map(x, ξ, d, k)
  sxx = S.(xx)
  sz = S.(z)

  s = :sgibbs
  dg1d.reset_session(s)
  dg1d.start_session(s)
  dg1d.cmd_session(cfg, s)
  dg1d.cmd_session(arrows, s)

  for i = 1:Npts
    lxx = map(x->dg1d.lagrange_poly(i, x, sz), sxx)
    dg1d.plot(hcat(xx, lxx), plotopts="w l title 'i = $i'", session=s)
  end

  dg1d.end_session(s)

  s = :sgibbs_fullgrid
  dg1d.reset_session(s)
  dg1d.start_session(s)
  dg1d.cmd_session(cfg, s)

  arrows *= repeat("unset arrow;", length(z))
  for zi in sz
    arrows *= "set arrow from $zi,-100.0 to $zi,100.0 nohead back;"
  end
  dg1d.cmd_session(arrows, s)

  for i = 1:Npts
    lxx = map(x->dg1d.lagrange_poly(i, x, sz), sxx)
    dg1d.plot(hcat(sxx, lxx), plotopts="w l title 'i = $i'", session=s)
  end

  dg1d.end_session(s)

  return
end

end
