export estimate_convergence_rate,
       purge_zeros!,
       least_sqr_linear,
       absolute_maximum,
       fd_approx,
       periodize,
       periodic_index


"""
    estimate_convergence_rate(n1, n2, f1, f2)

Experimental rate of convergence based on resolutions `n1, n2` and data `f1, f2`.
"""
function estimate_convergence_rate(n1, n2, f1, f2)
  return log10(f1/f2) / log10(n2/n1)
end


"""
    purge_zeros!(M, thrs=1e-13)

Scan `M` for elements smaller than `thrs` and set them to 0.
"""
purge_zeros!(M, thrs=1e-13) = M[abs.(M).<thrs] .= 0.0


"""
    least_sqr_linear(x, y)

Compute a linear least squares fit for data `x, y`.
Returns fit function `f, a, b`, where `f(x) = a + b * x` and `a,b` are the fitting coefficients.
Alogrithm taken from https://mathworld.wolfram.com/LeastSquaresFitting.html
"""
function least_sqr_linear(x, y)

  n = length(x)
  @assert n == length(y) "Mismatch in number of data points x, y!"

  yx = dot(y, x)
  xx = dot(x, x)
  mean_y = sum(y) / n
  mean_x = sum(x) / n

  ss_xx = xx - mean_x^2 * n
  ss_xy = yx - mean_x * mean_y * n

  b = ss_xy / ss_xx
  a = mean_y - b * mean_x

  f(x) = a + b * x
  return f, a, b
end


"""
    periodize(i, N)

Wrap `i` around period `N` if `i > N`, otherwise return `i`.
Behaves similar when called with `Int`s or `Real`s.
"""
@inline function periodize(i::T, N::S) where {T<:Int,S<:Int}
  mod = i % N
  return mod == 0 ? N : mod
end
@inline function periodize(x::T, L::S) where {T<:Real,S<:Real} 
  xround = floor(x/L) * L
  return x <= xround ? x : x - xround
end


@deprecate periodic_index(i, N) periodize(i, N)


@inline absolute_maximum(x) = norm(x, Inf)


"""
    fd_approx(f, r)

Finite difference approximation of first and second order derivatives.
Result is computed on grid points r[2:end-1], e.g. no derivative
is computed at domain endpoints
"""
function fd_approx(f, r)
  # TODO Add derivatives at endpoints with one-sided FD formulae.
  dr = diff(r)
  fr = zeros(length(f)-2)
  frr = zeros(length(f)-2)
  for i in 1:length(fr)
    dm1 = dr[i]
    dp1 = dr[i+1]
    dpd = dm1 + dp1
    fr[i] = - dp1 / (dm1*dpd) * f[i] + (dp1 - dm1) / (dm1*dp1) * f[i+1] + dm1 / (dp1 * dpd) * f[i+2]
    frr[i] = 2.0 / (dm1*dpd) * f[i] - 2.0 / (dm1*dp1) * f[i+1] + 2.0 / (dp1 * dpd) * f[i+2]
  end
  return fr, frr
end


"""
    central_dx(f, x)

Compute second order centered finite difference approximation for data `f`
on an unequaly spaced grid `x`.
Implementation is based on a second order Lagrange inteprolating polynomial
that connects data points around each grid point.
Boundary points are treated with one-sided differences of the same order.
"""
function central_dx(f, x)

  N = length(f)
  if N <= 1
    error("Require at least two data points!")
  end

  df = zeros(N)
  dx = diff(x)

  for i = 2:N-1
    dxm    = dx[i-1]
    dxp    = dx[i]
    sumdx  = dxm + dxp
    df[i]  += - dxp / (dxm * sumdx) * f[i-1]
    df[i]  += (dxp - dxm) / (dxp * dxm) * f[i]
    df[i]  += dxm / (dxp * sumdx) * f[i+1]
  end

  # forward/left-directed difference
  dxm = dx[1]
  dxp = dx[2]
  sumdx = dxm + dxp
  df[1] += - (2 * dxm + dxp) / (sumdx * dxm) * f[1]
  df[1] += sumdx / (dxm * dxp) * f[2]
  df[1] += - dxm / (dxp * sumdx) * f[3]

  # forward/left-directed difference
  dxm = dx[end-1]
  dxp = dx[end]
  sumdx = dxm + dxp
  df[N] += dxp / (dxm * sumdx) * f[N-2]
  df[N] += - sumdx / (dxp * dxm) * f[N-1]
  df[N] += (2 * dxp + dxm) / (sumdx * dxp) * f[N]

  return df
end


module FD

using dg1d
@deprecate central_dx(f, x) dg1d.central_dx(f, x)

end
