module EquationOfState


using dg1d
import Parameters: @unpack


export AbstractEquationOfState, Polytrope, IdealGas,
       Derivative, Density, Pressure, Enthalpy, TotalEnergy, InternalEnergy,
       SpeedOfSound, SquareSpeedOfSound, Enthalpy


#######################################################################
#                                Types                                #
#######################################################################


"""
    AbstractEquationOfState

Supertype for all Equations of State. To be used with the following method
definitions for the computation of thermodynamic variables `<:TVariable`:

```julia
function (eos::AbstractEquationOfState)(
    ::Type{TVariable_to_compute},
    ::Type{TVariable_1}, val1, ::Type{TVariable_2}, val2)

  # compute TVariable_to_compute from val1, val2
end
```

E.g. all equations of state are assumed to depend on two parameters. 
Allowed pairs of argument types `(TVariable_1, TVariable_2)` are
- `(Density, InternalEnergy)`, 
- `(Density, Pressure)`, 
- `(InternalEnergy, Pressure)`,
and always in alphabetic order.

# Examples

```jldoctest setup
julia> eos = IdealGas(gamma=1.4);

julia> rho = eos(Density, InternalEnergy, 1.0, Pressure, 1.0) 
2.5

julia> e = eos(TotalEnergy, Density, 1.0, InternalEnergy, 1.0) 
2.0
```
"""
abstract type AbstractEquationOfState end


"""
    TVariable

Supertype for all theromdynamic variables.
TVariables participate in the `EquationOfState` interface.
"""
abstract type TVariable end


Base.broadcastable(eos::AbstractEquationOfState) = Ref(eos)
Base.broadcastable(var::TVariable) = Ref(var)


#######################################################################
#                  Available thermodynamic variables                  #
#######################################################################


"""
Denoted by p.
"""
struct Pressure       <: TVariable end


"""
Denoted by ρ.
"""
struct Density        <: TVariable end


"""
Per unit mass, denoted by ϵ.
"""
struct InternalEnergy <: TVariable end


"""
Per unit mass, denoted by s.
"""
struct Entropy        <: TVariable end


"""
    Derivative{T<:TVariable,S<:TVariable} <: TVariable

TVariable for requesting the first derivative `dT/dS`.

===

# Examples

```jldoctest setup
julia> parameters = parse_Parameters(IdealGas);

julia> eos = IdealGas(parameters...);

julia> drhodp = eos(Derivative{Pressure,Density}, Density, 1.0, Pressure, 2.0)
0.8
```

See also: TVariable
"""
struct Derivative{T<:TVariable, S<:TVariable} <: TVariable end


#### Derived variables


"""
Per unit mass, denoted by e and defined as

  e = ρ (1 + ϵ).

"""
struct TotalEnergy    <: TVariable end


"""
Per unit mass, denoted by h.

  h = 1 + ϵ + p / ρ
"""
struct Enthalpy       <: TVariable end


"""
Denoted by cs.
  
  cs = sqrt((χ + (p / ρ^2) κ) / h)
"""
struct SpeedOfSound   <: TVariable end


"""
Denoted by cs^2.
  
  cs^2 = (χ + (p / ρ^2) κ) / h
"""
struct SquareSpeedOfSound   <: TVariable end




#######################################################################
#                           Implementations                           #
#######################################################################


include("derived_variables.jl")
include("polytrope.jl")
include("idealgas.jl")


function make_EquationOfState(prms)
  type = Symbol(query(prms, "type", default=missing))
  type === :missing && error("No equation of state provided!")

  concretes = [ Base.typename(ct).name for ct in concretetypes(AbstractEquationOfState) ]
  @assert type in concretes "'EquationOfState.type = $type' unknown. Available options: $concretes"
  eos_type = eval(type)
  eos_prms = parse_parameters(eos_type, prms)
  eos = eos_type(; eos_prms...)
  return eos
end


end
