"""
    IdealGas <: AbstractEquationOfState

Ideal gas law equation of state

  p(ρ,ϵ) = (Γ - 1) ρ ϵ

- `p` pressure
- `ϵ` internal energy
- `Γ` ratio of specific heats
"""
Base.@kwdef struct IdealGas <: AbstractEquationOfState
  gamma::Float64
end


#######################################################################
#                             Parameters                              #
#######################################################################


function dg1d.parse_parameters(::Type{IdealGas}, prms=Dict())
  gamma = query(Float64, prms, "gamma", default = 1.4)
  @assert gamma > 0
  return (; gamma)
end


#######################################################################
#                       Thermodynamic variables                       #
#######################################################################


# p = p(ρ,ϵ)
function (eos::IdealGas)(::Type{Pressure}, 
                         ::Type{Density}, rho, ::Type{InternalEnergy}, eps)

  return (eos.gamma - 1.0) * rho * eps
end


# ρ = ρ(ϵ,p)
function (eos::IdealGas)(::Type{Density}, 
                         ::Type{InternalEnergy}, eps, ::Type{Pressure}, p)

  return p / ((eos.gamma - 1.0) * eps)
end


# ϵ = ϵ(ρ,p)
function (eos::IdealGas)(::Type{InternalEnergy},
                         ::Type{Density}, rho, ::Type{Pressure}, p)
  
  return p / ((eos.gamma - 1.0) * rho)
end


# s = s(ρ,ϵ)
function (eos::IdealGas)(::Type{Entropy},
                         ::Type{Density}, rho, ::Type{InternalEnergy}, eps)

  return log(eps / rho^(eos.gamma - 1.0))
end


# χ = ∂p/∂ρ(ρ,ϵ)
function (eos::IdealGas)(::Type{Derivative{Pressure, Density}},
                         ::Type{Density}, rho, ::Type{InternalEnergy}, eps)

  return (eos.gamma - 1.0) * eps
end


# κ = ∂p/∂ϵ(̢ρ,ϵ)
function (eos::IdealGas)(::Type{Derivative{Pressure, InternalEnergy}},
                         ::Type{Density}, rho, ::Type{InternalEnergy}, eps) 

  return (eos.gamma - 1.0) * rho
end
