function analyze_evolution(filename)

  filename = abspath(filename)
  outputdir = dirname(filename)
  if !ispath(outputdir)
    mkdir(outputdir)
  end
  plotname(basename, nr) = joinpath(outputdir, "$(basename).$(lpad(nr,3,'0')).png")

  hist = nothing
  jldopen(filename, "r") do file
    hist = read(file, "hist")
  end

  prog = Progress(length(hist))

  for (i, h) in enumerate(hist)
    t, rhs = h
    plt = dg1d.plot(rhs.cons, rhs.mesh)
    Plots.plot!(plt, title="T = $t")
    pltname = plotname("conservatives", i)
    png(pltname)
    plt = dg1d.plot(rhs.scons, rhs.mesh)
    Plots.plot!(plt, title="T = $t")
    pltname = plotname("scaledconservatives", i)
    png(pltname)
    savefig(plt, pltname)
    plt = dg1d.plot(rhs.prim, rhs.mesh)
    Plots.plot!(plt, title="T = $t")
    pltname = plotname("primitives", i)
    png(pltname)
    next!(prog)
  end

end


function analyze_density(filename)

  filename = abspath(filename)
  outputdir = dirname(filename)
  if !ispath(outputdir)
    mkdir(outputdir)
  end

  hist = nothing
  jldopen(filename, "r") do file
    hist = read(file, "hist")
  end

  central_rho = Float64[]
  max_rho     = Float64[]
  ts          = Float64[]

  for h in hist
    t, data = h
    push!(ts, t)
    # depending on how data was stored we have to extract it differently
    # settle on output format/structure to avoid this
    rho = if data isa dg1d.GRHD.Rhs
      vec(data.prim.ρ)
    elseif data isa dg1d.GRHD.Primitives
      vec(data.ρ)
    end
    N = length(rho)
    if isodd(N)
      push!(central_rho, rho[Int64(N+1)/2])
    else
      push!(central_rho, (rho[Int64(N/2)] + rho[Int64(N/2)+1]) / 2)
    end
    push!(max_rho, maximum(rho))
  end

  # convert time data into milliseconds
  ts .*= dg1d.units.T * 1000
  # normalize data
  central_rho_rho0 = central_rho ./ central_rho[1] 
  max_rho_rho0     = max_rho ./ max_rho[1]

  # plot formatting
  yticks = (0.970:0.005:1.015)
  ylim = (minimum(yticks), maximum(yticks))
  yformatter = yi -> @sprintf("%.3f", yi)
  xlabel = "t/ms"

  plt = Plots.plot(title="Analysis rest mass density evolution")
  Plots.plot!(plt, ts, central_rho_rho0, label="ρ(t,x=0)/ρ(0,x=0)",
              yticks=yticks, ylim=ylim, yformatter=yformatter, xlabel=xlabel)
  Plots.plot!(plt, ts, max_rho_rho0, label="max[ρ(t)]/max[ρ(0)]",
              yticks=yticks, ylim=ylim, yformatter=yformatter, xlabel=xlabel,
              linestyle=:dash)
  plotname = joinpath(outputdir, "analysis_rho_rho0.pdf")
  @info "Exporting plot to '$plotname'"
  savefig(plt, plotname)

end
