@doc """
Formulae according to dg1d-notes/tex/grhd.tex
Notation:
  vr := v^r
  Sr := S_r
"""
module Formulae

  ### Auxilary functions

  # Lorentz factor
  Wprim(B, ρ, vr, ϵ, p) = 1.0 / sqrt(1.0 - B^2 * vr^2)
  Wcons(B, D, Sr, τ, p) = (τ + p + D) / sqrt((τ + p + D)^2 - (Sr / B)^2)
   
  ### Primitive variables
  # in terms of conservative ones

  ρ(B, D, Sr, τ, p)  = D / Wcons(B, D, Sr, τ, p)
  vr(B, D, Sr, τ, p) = Sr / (B^2 * (τ + D + p))
  ϵ(B, D, Sr, τ, p) = (sqrt((τ + p + D)^2 - (Sr/B)^2) - Wcons(B, D, Sr, τ, p) * p) / D - 1

  ### Conservative variables
  # in terms of primitive variables

  D(B, ρ, vr, ϵ, p) = ρ * Wprim(B, ρ, vr, ϵ, p)
  Sr(B, ρ, vr, ϵ, p) = ρ * (1 + ϵ + p / ρ) * Wprim(B, ρ, vr, ϵ, p)^2 * B^2 * vr
  τ(B, ρ, vr, ϵ, p) = ρ * (1 + ϵ + p / ρ) * Wprim(B, ρ, vr, ϵ, p)^2 - p - ρ * Wprim(B, ρ, vr, ϵ, p)

  ### Fluxes
  # in terms of primitive variables

  fD(B, ρ, vr, ϵ, p) = D(B, ρ, vr, ϵ, p) * vr
  fSr(B, ρ, vr, ϵ, p) = Sr(B, ρ, vr, ϵ, p) * vr + p
  fτ(B, ρ, vr, ϵ, p) = (τ(B, ρ, vr, ϵ, p) + p) * vr

  #### Sources

  function sD(r, A, A_r, B, B_r, ρ, vr, ϵ, p) 
    val = - A * B^3 * D(B, ρ, vr, ϵ, p) * vr
    return val
  end

  function sSr(r, A, A_r, B, B_r, ρ, vr, ϵ, p)
    h = 1 + ϵ + p / ρ
    Sr = Formulae.Sr(B, ρ, vr, ϵ, p)
    W = Wprim(B, ρ, vr, ϵ, p)
    v2 = B^2 * vr^2
    ρhW2 = ρ * h * W^2
    val = - (ρhW2 - p) * r * B^3 * A_r
    val += A * B^4 * (r * B_r + B / 3) * (ρhW2 * vr^2 + 3 * p  / B^2)
    val += - 4 / 3 * A * B^3 * Sr * vr
    return val
  end

  function sτ(r, A, A_r, B, B_r, ρ, vr, ϵ, p)
    τ = Formulae.τ(B, ρ, vr, ϵ, p)
    h = 1 + ϵ + p / ρ
    W = Wprim(B, ρ, vr, ϵ, p)
    ρhW2 = ρ * h * W^2
    val = - r * B^3 * ρhW2 * vr * A_r - A * B^3 * (τ + p) * vr
    return val
  end

end


function find_pressure(eos::Polytrope, p0, B, D, Sr, τ, r;
    ρatm=1e-10, atm_threshold = 100.0,
    prms::dg1d.NlSolveParameters=dg1d.Default_NlSolveParameters)

  # upper limit on particle velocity
  omv2 = 1e-3 # = 1 - v^2

  S2 = (Sr/B)^2

  function sanitize_pressure(p)
    a2mS2 = (τ + D + p)^2 - S2
    if a2mS2 < 0.0
      p = - (τ + D) + sqrt(S2 + omv2) 
    end
    return p
  end

  p0 = sanitize_pressure(p0)

  # functor for root search
  function f_and_df(p)

    p = sanitize_pressure(p)

    a = τ + p + D
    a2mS2 = a^2 - S2
    W = Formulae.Wcons(B, D, Sr, τ, p)
    ρ = D / W

    # apply atmosphere treatment
    if ρ < atm_threshold * ρatm
      ρ = ρatm
    end

    # P = NewEoS.compute(Pressure, eos, Density(ρ))
    # dPdρ = NewEoS.compute_derivative(Pressure, Density, eos, Density(ρ))
    P = eos(Pressure, Density(ρ))
    dPdρ = eos(Derivative{Pressure, Density}, Density(ρ))
    dρdp = D * S2 * W / a^3

    # residual and its derivative
    f = P - p
    df = dPdρ * dρdp - 1

    return f, df
  end

  p, err, converged = dg1d.nlsolve_exact(f_and_df, p0, prms=prms)

  if !converged
    println()
    error("Could not find pressure `p` at r = $r that is consistent with equation of state!")
  end

  return p
end


function impose_atmosphere!(rhs::GrhdEquations; ρatm=ρatm, atm_threshold=atm_threshold)

  @unpack eos, primitives = rhs
  @unpack conservatives = rhs.balancelaw

  patm = eos(Pressure, Density(ρatm))
  hatm = eos(Enthalpy, Pressure(patm), Density(ρatm))
  ϵatm = eos(InternalEnergy, Pressure(patm))
  Watm = 1.0

  for (i,ρi) in enumerate(primitives.ρ)
    if ρi < ρatm * atm_threshold
      primitives.ρ[i] = ρatm
      primitives.p[i] = patm
      primitives.ϵ[i] = ϵatm
      primitives.vr[i] = 0.0
      conservatives.D[i] = ρatm * Watm
      conservatives.Sr[i] = 0.0
      conservatives.τ[i] = ρatm * hatm * Watm^2 - patm - ρatm * Watm
    end
  end

  return
end


function compute_primitives!(grhdeqs::GrhdEquations,
    ρatm=1e-10, atm_threshold = 100.0,
    newton_prms::dg1d.NlSolveParameters=dg1d.Default_NlSolveParameters)

  @unpack conservatives = grhdeqs.balancelaw
  @unpack eos, primitives, metricfns = grhdeqs
  @unpack D, Sr, τ = conservatives
  @unpack ρ, vr, p, ϵ = primitives
  @unpack B = metricfns

  @. p = find_pressure(eos, p, B, D, Sr, τ, vec(mesh.x),
                        ρatm=ρatm, atm_threshold=atm_threshold)
  @. ρ = Formulae.ρ(B, D, Sr, τ, p)
  @. vr = Formulae.vr(B, D, Sr, τ, p)
  @. ϵ = Formulae.ϵ(B, D, Sr, τ, p)

  impose_atmosphere!(rhs, ρatm=ρatm, atm_threshold=atm_threshold)

  return
end


function compute_conservatives!(cons::Conservatives, prim::Primitives, mfs::MetricFns)
  @unpack D, Sr, τ = cons
  @unpack ρ, vr, p, ϵ = prim
  @unpack B = mfs

  @. D = Formulae.D(B, ρ, vr, ϵ, p)
  @. Sr = Formulae.Sr(B, ρ, vr, ϵ, p)
  @. τ = Formulae.τ(B, ρ, vr, ϵ, p)

  return
end


function compute_fluxes!(fluxes::Fluxes, prim::Primitives, mfs::MetricFns, r)
  @unpack fD, fSr, fτ = fluxes
  @unpack ρ, vr, p, ϵ = prim
  @unpack A, B = mfs

  @. fD = r * A * B^3 * Formulae.fD(B, ρ, vr, ϵ, p)
  @. fSr = r * A * B^3 * Formulae.fSr(B, ρ, vr, ϵ, p)
  @. fτ = r * A * B^3 * Formulae.fτ(B, ρ, vr, ϵ, p)

  return
end


function compute_ldgfluxes!(ldgfluxes::LdgFluxes, ldgq::LdgVariables, μ)

  vars_ldgf = fields(ldgfluxes)
  vars_ldgq = fields(ldgq)

  for (f, q) in zip(vars_ldgf, vars_ldgq)
    @. f = μ[:] * q
  end

  return
end


function compute_sources!(sources::Sources, prim::Primitives, mfs::MetricFns, r;
    ρatm=1e-10, atm_threshold = 100.0)


  @unpack sD, sSr, sτ = sources
  @unpack ρ, vr, p, ϵ = prim
  @unpack A, A_r, B, B_r = mfs

  @. sD = Formulae.sD(r, A, A_r, B, B_r, ρ, vr, ϵ, p)
  @. sSr = Formulae.sSr(r, A, A_r, B, B_r, ρ, vr, ϵ, p)
  @. sτ = Formulae.sτ(r, A, A_r, B, B_r, ρ, vr, ϵ, p)

  for (i,ρi) in enumerate(prim.ρ)
    if ρi < ρatm * atm_threshold
      sD[i]   = 0.0
      sSr[i]  = 0.0
      sτ[i]   = 0.0
    end
  end

  return
end


function compute_conservatives!(cons::Conservatives, scons::ScaledConservatives,
    mfs::MetricFns, mesh::Mesh)
  @unpack x, K, element = mesh
  @unpack Npts = element

  if isodd(Npts) && isodd(K)
    error("Cannot yet handle case where r = 0 falls into symmetric cell with even polynomial!")
  end

  # find cells where x coordinate is zero
  # firstly, find indices of mesh points that are zero
  idxs = findall(xi->abs(xi) < 1e-14, mesh.x)

  # apply transformation blindly to all points
  @unpack B = mfs
  for (var_con, var_scon) in zip(fields(cons), fields(scons))
    @. var_con = var_scon / (B^3 * x[:])
  end

  # postprocess points where we divided by zero above
  # do so by interpolating value where x = 0 using symmetry/anti-symmetry
  sym_cons = ( cons.D, cons.τ )
  asym_cons = ( cons.Sr, )

  tmp = zeros(Npts)
  for idx in idxs
    i, j = Tuple(idx)
    for var_con in sym_cons
      mat_var_con = reshape(var_con, (Npts, K))
      tmp .= mat_var_con[:,j]
      tmp[i] = 0 # here we divided by zero
      Di = element.D[i,:]
      Dii = Di[i]
      Di[i] = 0.0
      tmp[i] = - dot(Di, tmp) / Dii
      mat_var_con[:,j] .= tmp
    end
    for var_con in asym_cons
      var_con[i+(j-1)*Npts] = 0
    end
  end

  # sanity check
  @assert isnan(cons) == false
  @assert isinf(cons) == false

  return
end


function compute_scaledconservatives!(scons::ScaledConservatives, cons::Conservatives,
    mfs::MetricFns, mesh::Mesh)
  @unpack x = mesh
  @unpack B = mfs
  for (var_con, var_scon) in zip(fields(cons), fields(scons))
    @. var_scon = var_con * B^3 * x[:]
  end
  return
end


#######################################################################
#                             Deprecated                              #
#######################################################################


function impose_atmosphere!(rhs::GRHD.Rhs; ρatm=ρatm, atm_threshold=atm_threshold)

  @unpack eos, cons, prim = rhs

  patm = eos(Pressure, Density(ρatm))
  hatm = eos(Enthalpy, Pressure(patm), Density(ρatm))
  ϵatm = eos(InternalEnergy, Pressure(patm))
  Watm = 1.0

  for (i,ρi) in enumerate(prim.ρ)
    if ρi < ρatm * atm_threshold
      prim.ρ[i] = ρatm
      prim.p[i] = patm
      prim.ϵ[i] = ϵatm
      prim.vr[i] = 0.0
      cons.D[i] = ρatm * Watm
      cons.Sr[i] = 0.0
      cons.τ[i] = ρatm * hatm * Watm^2 - patm - ρatm * Watm
    end
  end

  return
end


function compute_primitives!(rhs::Rhs; 
    ρatm=1e-10, atm_threshold = 100.0,
    newton_prms::dg1d.NlSolveParameters=dg1d.Default_NlSolveParameters)

  @unpack cons, prim, eos, mfs, mesh = rhs
  @unpack D, Sr, τ = cons
  @unpack ρ, vr, p, ϵ = prim
  @unpack B = mfs

  @. p = find_pressure(eos, p, B, D, Sr, τ, mesh.x[:],
                        ρatm=ρatm, atm_threshold=atm_threshold)
  @. ρ = Formulae.ρ(B, D, Sr, τ, p)
  @. vr = Formulae.vr(B, D, Sr, τ, p)
  @. ϵ = Formulae.ϵ(B, D, Sr, τ, p)

  impose_atmosphere!(rhs, ρatm=ρatm, atm_threshold=atm_threshold)

  return
end
