function find_troubled_cells(fn, var, mesh)
  idxs = findall(fn, var)
  cells = [ Tuple(idx)[2] for idx in idxs ]
  unique!(cells)
  return cells
end


function reconstruct_primitives!(
    grhdeqs::GrhdEquations, bernstein_rec::dg1d.BernsteinReconstruction, ρatm; 
    reconstruction_variables::Vector{Symbol}=[:vr], impose_bounds::Bool=true)

  variables = grhdeqs.primitives
  vars = fieldnames(typeof(variables))
  @assert subset(tuple(reconstruction_variables...), vars) # "Provided variables '$vars' not available in `$vars`"
  # TODO Check if this assertion still makes sense
  # @assert all(in.(reconstruction_variables, Ref(vars))) "Provided variables '$vars' not available in `$vars`"

  @unpack mesh = grhdeqs.balancelaw
  shape = layout(mesh)
  Npts, K = shape

  # # troubled cells based on where visocisty is non-vanishing
  # @unpack μ = rhs.av
  # indicator = dg1d.TroubledCells.ArtificialViscosityIndicator(0.8)
  # μnorm = sum(μ, dims=1) / Npts
  # μnorm_max = maximum(μnorm)
  # if μnorm_max > 0.0
  #   μnorm ./= μnorm_max
  # end
  #
  # trbld_cells = dg1d.TroubledCells.find_troubled_cells(indicator, μnorm, mesh)
  trbld_cells = grhdeqs.troubledcells
  push!(trbld_cells, 1)
  push!(trbld_cells, K)
  unique!(trbld_cells)

  # use only Bernstein reconstruction for now
  w = 0.0 # convex weighting

  # atmosphere values
  patm = rhs.eos(Pressure, Density(ρatm))
  ϵatm = rhs.eos(InternalEnergy, Pressure(patm))
  vmax = 1.0 - 1e-3

  # apply reconstruction to selected variables `vars`
  for rv in reconstruction_variables
    var = getfield(variables, rv)
    mat_var = reshape(var, shape)
    for cell in trbld_cells
      trbld_var = view(mat_var, 1:Npts, cell)
      dg1d.reconstruct!(trbld_var, w, bernstein_rec)
      # if impose_bounds
      #   if :ρ in vars
      #     dg1d.reconstruct!(trbld_var, w, bernstein_rec, m=ρatm)
      #   elseif :p in vars
      #     dg1d.reconstruct!(trbld_var, w, bernstein_rec, m=patm)
      #   elseif :ϵ in vars
      #     dg1d.reconstruct!(trbld_var, w, bernstein_rec, m=ϵatm)
      #   elseif :vr in vars
      #     dg1d.reconstruct!(trbld_var, w, bernstein_rec, m=-vmax, M=vmax)
      #   end
      # else
      #   dg1d.reconstruct!(trbld_var, w, bernstein_rec)
      # end
    end
  end

end


function new_rhs_bernstein!(rhs::Rhs, dU, U, t, 
    components, ρatm, atm_threshold, 
    bernstein_rec::dg1d.BernsteinReconstruction,
    grhdeqs::GrhdEquations)

  @unpack mesh = rhs
  @unpack K = mesh
  @unpack Npts = mesh.element
  shape = (Npts, K)

  # Rs = 8.125144892166112
  # idx_lhs_surface = dg1d.locate_point(-Rs, mesh)
  # idx_rhs_surface = dg1d.locate_point(Rs, mesh)
  # stencile_width = 4
  # stencile(i) = (i-stencile_width):(i+stencile_width)
  # rng_lhs_surface = (:, stencile(idx_lhs_surface[2]))
  # rng_rhs_surface = (:, stencile(idx_lhs_surface[2]))

  ### interpret inputs
  sv_u = StateVector(U, components)
  sv_du = StateVector(dU, components)
  rhs.scons = alias_component(sv_u, ScaledConservatives)
  d_scons = alias_component(sv_du, ScaledConservatives)

  @unpack conservatives, primitives, metricfns, scaledconservatives = grhdeqs
  compute_conservatives!(conservatives, scaledconservatives, metricfns, mesh)
  compute_primitives!(grhdeqs, ρatm=ρatm, atm_threshold=atm_threshold)
  reconstruct_primitives!(grhdeqs, bernstein_rec, ρatm, reconstruction_variables=[:vr],
    impose_bounds=false)
  compute_conservatives!(rhs.cons, rhs.prim, rhs.mfs)
  compute_scaledconservatives!(rhs.scons, rhs.cons, rhs.mfs, mesh)

  compute_fluxes!(rhs.fluxes, rhs.prim, rhs.mfs, mesh.x[:])
  compute_sources!(rhs.sources, rhs.prim, rhs.mfs, mesh.x[:];
    ρatm=ρatm, atm_threshold=atm_threshold)

  # numerical fluxes

  max_v = 1.0

  ### actual RHS computation
  # solve conservation law: ∂t u + ∂x f(u) + ∂x μ q = s

  @unpack fD, fSr, fτ = rhs.fluxes
  @unpack hatD, hatSr, hatτ = rhs.scons

  # numerical fluxes
  # For now we use speed of light as upper bound for lax friedrich flux
  max_v = 1.0
  # TODO implement estimation of max wave speed, requires characteristics
  # max_v = maxabs_v(eq, ...)

  intf_hatD = dg1d.to_interface(reshape(hatD, shape), mesh)
  intf_hatSr = dg1d.to_interface(reshape(hatSr, shape), mesh)
  intf_hatτ = dg1d.to_interface(reshape(hatτ, shape), mesh)
  intf_fD = dg1d.to_interface(reshape(fD, shape), mesh)
  intf_fSr = dg1d.to_interface(reshape(fSr, shape), mesh)
  intf_fτ = dg1d.to_interface(reshape(fτ, shape), mesh)
  nfD_lhs, nfD_rhs = dg1d.lax_friedrich_flux(intf_fD, intf_hatD, max_v)
  nfSr_lhs, nfSr_rhs = dg1d.lax_friedrich_flux(intf_fSr, intf_hatSr, max_v)
  nfτ_lhs, nfτ_rhs = dg1d.lax_friedrich_flux(intf_fτ, intf_hatτ, max_v)
  # nfD_lhs, nfD_rhs = dg1d.central_flux(intf_fD)
  # nfSr_lhs, nfSr_rhs = dg1d.central_flux(intf_fSr)
  # nfτ_lhs, nfτ_rhs = dg1d.central_flux(intf_fτ)

  # # Set boundary conditions
  # Not needed for now, because we try periodic BCs first.
  # nfD_lhs, nfD_rhs
  # nfSr_lhs, nfSr_rhs
  # nfτ_lhs, nfτ_rhs

  dg1d.new_rhs_weak!(d_scons.hatD, fD, rhs.sources.sD, nfD_lhs, nfD_rhs, mesh)
  dg1d.new_rhs_weak!(d_scons.hatSr, fSr, rhs.sources.sSr, nfSr_lhs, nfSr_rhs, mesh)
  dg1d.new_rhs_weak!(d_scons.hatτ, fτ, rhs.sources.sτ, nfτ_lhs, nfτ_rhs, mesh)

  # needed to mitigate effects caused by imposing periodic boundary conditions wrongly
  d_scons.hatD[1] = 0.0; d_scons.hatD[end] = 0.0
  d_scons.hatSr[1] = 0.0; d_scons.hatSr[end] = 0.0
  d_scons.hatτ[1] = 0.0; d_scons.hatτ[end] = 0.0

  return
end
