function rhs_ldg!(rhs::Rhs, dU, U, t, components, ρatm, atm_threshold)
  # @show t

  @unpack mesh = rhs
  @unpack K = mesh
  @unpack Npts = mesh.element
  shape = (Npts, K)

  rng = (:, 67:69)

  ### interpret inputs
  sv_u = StateVector(U, components)
  sv_du = StateVector(dU, components)
  rhs.scons = get_component(1, sv_u)
  d_scons = get_component(1, sv_du)

  ### compute fields
  compute_conservatives!(rhs.cons, rhs.scons, rhs.mfs, mesh)
  compute_primitives!(rhs, ρatm=ρatm, atm_threshold=atm_threshold)


  # display(dg1d.Segment(rhs.scons, mesh, rng...))
  # display(dg1d.Segment(rhs.cons, mesh, rng...))
  # display(dg1d.Segment(rhs.prim, mesh, rng...))
  #
  # @debug_plot(rhs.prim, mesh)

  compute_fluxes!(rhs.fluxes, rhs.prim, rhs.mfs, mesh.x[:])
  # compute_sources!(rhs.sources, rhs.prim, rhs.mfs, mesh.x[:])
  compute_sources!(rhs.sources, rhs.prim, rhs.mfs, mesh.x[:];
    ρatm=ρatm, atm_threshold=atm_threshold)

  ### LDG computation
  # solve auxilary equation: q + ∂x u = 0

  @unpack hatD, hatSr, hatτ = rhs.scons
  @unpack qD, qSr, qτ = rhs.ldgq

  # using rhs.ldgfluxes as temporary storage for filtered scons variables
  for (ldgf, var) in zip(fields(rhs.ldgfluxes), fields(rhs.scons))
    mat_var = reshape(var, shape)
    mat_ldgf = reshape(ldgf, shape)
    mat_ldgf .= mat_var
  end

  # numerical fluxes
  @unpack gD, gSr, gτ = rhs.ldgfluxes
  intf_gD = dg1d.to_interface(reshape(gD, shape), mesh)
  intf_gSr = dg1d.to_interface(reshape(gSr, shape), mesh)
  intf_gτ = dg1d.to_interface(reshape(gτ, shape), mesh)
  ngD_lhs, ngD_rhs = dg1d.central_flux(intf_gD)
  ngSr_lhs, ngSr_rhs = dg1d.central_flux(intf_gSr)
  ngτ_lhs, ngτ_rhs = dg1d.central_flux(intf_gτ)

  # # Set boundary conditions
  # Not needed for now, because we try periodic BCs first.

  # rhs
  dg1d.new_rhs_weak!(qD, gD, ngD_lhs, ngD_rhs, mesh)
  dg1d.new_rhs_weak!(qSr, gSr, ngSr_lhs, ngSr_rhs, mesh)
  dg1d.new_rhs_weak!(qτ, gτ, ngτ_lhs, ngτ_rhs, mesh)

  # compute LDG associated fluxes, needed for LDG regularization below
  # TODO Promote μ to AbstractVariable 
  compute_ldgfluxes!(rhs.ldgfluxes, rhs.ldgq, rhs.av.μ)

  ### actual RHS computation
  # solve conservation law: ∂t u + ∂x f(u) + ∂x μ q = s

  @unpack fD, fSr, fτ = rhs.fluxes
  @unpack hatD, hatSr, hatτ = rhs.scons
  @unpack qD, qSr, qτ = rhs.ldgq
  @unpack gD, gSr, gτ = rhs.ldgfluxes

  # # Set boundary conditions
  # Not needed for now, because we try periodic BCs first.
  # nqD_lhs, nqD_rhs
  # nqSr_lhs, nqSr_rhs
  # nqτ_lhs, nqτ_rhs

  # numerical fluxes
  # For now we use speed of light as upper bound for lax friedrich flux
  max_v = 1.0
  # TODO implement estimation of max wave speed, requires characteristics
  # max_v = maxabs_v(eq, ...)

  intf_hatD = dg1d.to_interface(reshape(hatD, shape), mesh)
  intf_hatSr = dg1d.to_interface(reshape(hatSr, shape), mesh)
  intf_hatτ = dg1d.to_interface(reshape(hatτ, shape), mesh)
  intf_fD = dg1d.to_interface(reshape(fD, shape), mesh)
  intf_fSr = dg1d.to_interface(reshape(fSr, shape), mesh)
  intf_fτ = dg1d.to_interface(reshape(fτ, shape), mesh)
  nfD_lhs, nfD_rhs = dg1d.lax_friedrich_flux(intf_fD, intf_hatD, max_v)
  nfSr_lhs, nfSr_rhs = dg1d.lax_friedrich_flux(intf_fSr, intf_hatSr, max_v)
  nfτ_lhs, nfτ_rhs = dg1d.lax_friedrich_flux(intf_fτ, intf_hatτ, max_v)
  # nfD_lhs, nfD_rhs = dg1d.central_flux(intf_fD)
  # nfSr_lhs, nfSr_rhs = dg1d.central_flux(intf_fSr)
  # nfτ_lhs, nfτ_rhs = dg1d.central_flux(intf_fτ)

  intf_gD = dg1d.to_interface(reshape(gD, shape), mesh)
  intf_gSr = dg1d.to_interface(reshape(gSr, shape), mesh)
  intf_gτ = dg1d.to_interface(reshape(gτ, shape), mesh)
  ngD_lhs, ngD_rhs = dg1d.central_flux(intf_gD)
  ngSr_lhs, ngSr_rhs = dg1d.central_flux(intf_gSr)
  ngτ_lhs, ngτ_rhs = dg1d.central_flux(intf_gτ)
  # intf_qD = dg1d.to_interface(reshape(qD, shape), mesh)
  # intf_qSr = dg1d.to_interface(reshape(qSr, shape), mesh)
  # intf_qτ = dg1d.to_interface(reshape(qτ, shape), mesh)
  # ngD_lhs, ngD_rhs = dg1d.central_flux(intf_qD)
  # ngSr_lhs, ngSr_rhs = dg1d.central_flux(intf_qSr)
  # ngτ_lhs, ngτ_rhs = dg1d.central_flux(intf_qτ)

  # # Set boundary conditions
  # Not needed for now, because we try periodic BCs first.
  # nfD_lhs, nfD_rhs
  # nfSr_lhs, nfSr_rhs
  # nfτ_lhs, nfτ_rhs

  dg1d.new_rhs_weak!(d_scons.hatD, fD + gD, rhs.sources.sD, nfD_lhs + ngD_lhs, nfD_rhs + ngD_rhs, mesh)
  dg1d.new_rhs_weak!(d_scons.hatSr, fSr + gSr, rhs.sources.sSr, nfSr_lhs + ngSr_lhs, nfSr_rhs + ngSr_rhs, mesh)
  dg1d.new_rhs_weak!(d_scons.hatτ, fτ + gτ, rhs.sources.sτ, nfτ_lhs + ngτ_lhs, nfτ_rhs + ngτ_rhs, mesh)
  # dg1d.new_rhs_weak!(d_scons.hatD, fD, rhs.sources.sD, nfD_lhs, nfD_rhs, mesh)
  # dg1d.new_rhs_weak!(d_scons.hatSr, fSr, rhs.sources.sSr, nfSr_lhs, nfSr_rhs, mesh)
  # dg1d.new_rhs_weak!(d_scons.hatτ, fτ, rhs.sources.sτ, nfτ_lhs, nfτ_rhs, mesh)
  # dg1d.new_rhs_weak!(d_scons.hatD, gD, ngD_lhs, ngD_rhs, mesh)
  # dg1d.new_rhs_weak!(d_scons.hatSr, gSr, ngSr_lhs, ngSr_rhs, mesh)
  # dg1d.new_rhs_weak!(d_scons.hatτ, gτ, ngτ_lhs, ngτ_rhs, mesh)

  return
end
