function initialdata!(rhs::RHS, prms)
  initialdata_equation!(rhs, rhs.equation, prms)
  initialdata_hrsc!(rhs, rhs.hrsc, prms)
  initialdata_tci!(rhs, rhs.tci, prms)
end


function initialdata_equation!(rhs::RHS, equation::AbstractScalarEquation, parameters)
  @unpack cache, mesh = rhs
  @unpack u = get_dynamic_variables(cache)
  @unpack L, x = mesh
  x = vec(x)
  tstart = 0.0 # always
  id_fn = make_function_initialdata_equation(parameters)
  @. u = id_fn(x)
  return
end


"""
    make_function_initialdata(parameters)

Generate a function `id_fn(x)` to compute initial data on grid `x`.
`parameters` must contain sections `Initialdata, Mesh`.
"""
function make_function_initialdata_equation(parameters)

  type = query(parameters, "type", options=["sine", "bump", "bugner gaussian", "gaussian"])
  xrange = query(Vector{Float64}, parameters, "xrange", default = [-1.0,1.0])
  L = first(diff(xrange))

  id_fn = if type == "sine"
    period = query(Int64, parameters, "period", default=1) * 2.0 * π / L
    phase = query(Float64, parameters, "phase", default=0.0)
    (x) -> sin(period * x + phase)
  elseif type == "bump"
    xmin, xmax = xrange
    height = query(Float64, parameters, "height", default=1)
    width = query(Float64, parameters, "width", default=(xmax-xmin)/2)
    @assert width < L "Initialdata 'bump': Bump width exceeds x domain '$xrange'"
    xl, xr = xmin+(L-width)/2, xmin+(L+width)/2
    (x) -> (x < xl || x > xr) ? 0.5 : height
  elseif type == "gaussian"
    amplitude = query(Float64, parameters, "height", default=1)
    width = query(Float64, parameters, "width", default=1)
    x0 = (xrange[2] - xrange[1]) / 2
    (x) -> amplitude * exp(-(x-x0)^2 / width^2) + 0.2
  elseif type == "bugner gaussian"
    xmin, xmax = xrange
    @assert xmin == -3.0 && xmax == 3.0
    A, sigma = 1.0, 0.4
    (x) -> A * exp(-x^2 / sigma^2)
  elseif type == "smooth bump"
    error("TODO Implement 'smooth bump' initial data")
#   elseif type === :smooth_bump
#     xl = xmin + L * 0.10
#     xr = xmin + L * 0.90
#     gap_x = 0.3 * L
#     u0 = zeros(size(x))
#     for (i, xi) in enumerate(x)
#       if xi < xl || xi > xr
#         u0[i] = 0.0
#       elseif xi > xl+gap_x && xi < xr-gap_x
#         u0[i] = 1.0
#       elseif xi >= xl && xi <= xl+gap_x
#         u0[i] = sin( (xi - xl) / gap_x * 0.5 * π)^2
#       elseif xi >= xr-gap_x && xi <= xr
#         u0[i] = 1.0 - sin( (xi - (xr - gap_x)) / gap_x * 0.5 * π)^2
#       end
#     end
  else
    error("Requested initialdata type $id_type not implemented!")
  end

  return id_fn
end


function initialdata_hrsc!(rhs::RHS, hrsc::Maybe{HRSC.AbstractHRSC}, prms) end


function initialdata_hrsc!(rhs::RHS, hrsc::HRSC.AbstractArtificialViscosity, prms)
  @unpack mu = get_cell_variables(rhs.cache)
  @. mu = 0.0
end


function initialdata_hrsc!(rhs::RHS, hrsc::HRSC.EntropyViscosity, prms)
  @unpack E, Em1, F, Fm1 = get_static_variables(rhs.cache)
  @unpack mu             = get_cell_variables(rhs.cache)
  @. E = 0.0
  @. F = 0.0
  @. Em1 = 0.0
  @. Fm1 = 0.0
  @. mu = 0.0
end


function initialdata_tci!(rhs::RHS, tci::Nothing, prms) end


function initialdata_tci!(rhs::RHS, tci::TCI.AbstractTCI, prms)
  @unpack flag = get_cell_variables(rhs.cache)
  @. flag = 0.0
end
