Base.@kwdef struct Minmod <: AbstractTCI
  mesh::Mesh
  # parameters
  threshold::Float64
end


function parse_parameters(::Type{Minmod}, prms)
  threshold = query(Float64, prms, "threshold", default=1e-5)
  @assert threshold > 0
  return (; threshold)
end


function compute_indicator!(
    flag, # mutated outputs
    u,    # inputs
    tci::Minmod)

  @unpack mesh       = tci
  @unpack invjac, dl = mesh
  @unpack D          = mesh.element
  L                  = layout(mesh)
  Npts, K            = L

  fill!(flag, 0)
  u = reshape(u, L)
  u_averages = dg1d.cellwise_average(u, mesh)

  @unpack threshold = tci
  # fill flag by applying minmod criterion
  for k in 1:K # loop over cells
    # assuming periodic grid
    k_lhs, k_rhs = periodic_index(k-1, K), periodic_index(k+1, K)
    uavg, uavg_lhs, uavg_rhs = u_averages[k], u_averages[k_lhs], u_averages[k_rhs]
    ubdry_lhs, ubdry_rhs = u[1,k], u[end,k]

    delta_avg_lhs = uavg - uavg_lhs
    delta_avg_rhs = uavg_rhs - uavg
    delta_bdry_lhs = uavg - ubdry_lhs
    delta_bdry_rhs = ubdry_rhs - uavg

    # troubled check, cf. (3.64) in Bugner 2017, threshold =^= M
    if !(minmod_M(dl, threshold, delta_bdry_lhs, delta_avg_lhs, delta_avg_rhs) ≈ delta_bdry_lhs) ||
       !(minmod_M(dl, threshold, delta_bdry_rhs, delta_avg_lhs, delta_avg_rhs) ≈ delta_bdry_rhs)
      flag[k] = 1
    end
  end

  return
end


function minmod(args::T...)::T where T <: Real
  signs = collect(sign.(args))
  abs_signs = abs.(signs)
  if sum(abs_signs) == T(length(args))
    return signs[1] * minimum(abs.(args))
  else
    return T(0)
  end
end


function minmod_M(dx::T, M::T, args::T...)::T where T <: Real
  if abs(args[1]) <= M * dx^2
    return args[1]
  else
    return minmod(args...)
  end
end
