@doc raw"""
Compute RHS of weak DG approximation of a conservation law:
```math
  \partial_t u = - \partial_x f(u)
  \rightarrow
  u_i = M^{-1}_{ij} S^{jk} f_k 
        - M^{-1}_{ij} ( l_j(LHS) (nf)^{*}(LHS) + l_j(RHS) (nf)^{*}(RHS) )
```
"""
function rhs_weak!(du, f, nf_lhs, nf_rhs, mesh::Mesh)
  # extract operators
  @unpack invjac, element = mesh
  @unpack MDM, Ml_lhs, Ml_rhs = element
  # evaluate rhs
  mul!(du, MDM, f)
  du .-= (transpose(nf_rhs) .* Ml_rhs .- transpose(nf_lhs) .* Ml_lhs)
  # account for coordinate transformation onto physical grid
  du .*= invjac
end # function


@doc raw"""
Compute RHS of strong DG approximation of a conservation law:
```math
  \partial_t u = - \partial_x f(u)
  \rightarrow
  u_i = - D^{jk} u_k 
        + M^{-1}_{ij} ( l_j(LHS) (nf(LHS) - (nf)^{*}(LHS)) 
                      + l_j(RHS) (nf(RHS) - (nf)^{*}(RHS)) )
```
"""
function rhs_strong!(du, f, f_min_nf_lhs, f_min_nf_rhs, mesh::Mesh)
  # extract operators
  @unpack invjac, element = mesh
  @unpack D, Ml_lhs, Ml_rhs = element
  # evaluate rhs
  mul!(du, D, - f)
  du .+= (transpose(f_min_nf_rhs) .* Ml_rhs .-
          transpose(f_min_nf_lhs) .* Ml_lhs)
  # account for coordinate transformation onto physical grid
  du .*= invjac
end # function



function new_rhs_weak!(vec_du, vec_f, nf_lhs, nf_rhs, mesh::Mesh)
  # extract operators
  @unpack invjac, element = mesh
  @unpack Npts, MDM, Ml_lhs, Ml_rhs = element
  @unpack K = mesh
  du = reshape(vec_du, (Npts, K))
  f = reshape(vec_f, (Npts, K))
  # evaluate rhs
  mul!(du, MDM, f)
  du .-= (transpose(nf_rhs) .* Ml_rhs .- transpose(nf_lhs) .* Ml_lhs)
  # account for coordinate transformation onto physical grid
  du .*= invjac
  return
end # function



function new_rhs_weak!(vec_du, vec_f, vec_s, nf_lhs, nf_rhs, mesh::Mesh)
  @unpack invjac, element = mesh
  @unpack Npts, invM, MDM, Ml_lhs, Ml_rhs = element
  @unpack K = mesh
  du = reshape(vec_du, (Npts, K))
  f = reshape(vec_f, (Npts, K))
  s = reshape(vec_s, (Npts, K))
  mul!(du, MDM, f)
  @. du += - (nf_rhs' * Ml_rhs - nf_lhs' * Ml_lhs)
  # account for coordinate transformation onto physical grid
  du .*= invjac
  du .+= s
  return
end # function
