export Bisection, NewtonSolver, find_root


"""
    abstract type Rootfinder end

Super type for Rootfinder interface:


    MyRootfinder <: Rootfinder

Collect solver options. Must provide the following fields
- `reltol` is relative tolerance wrt residual of initial iterate, e.g. `reltol * abs(f(x_0))`,
- `abstol` is absolute tolerance,
- `maxits` is maximum number of iteration steps.


    find_root(f_df, x0_or_xrange, method::MyRootfinder; sanitize=identity)

Find root of function `f_df` which computes the residual and its derivative,
given an initial iterate or search interval `x0_or_xrange` and `method`.
`sanitize` is optional and can be used to clean up an iterate, e.g. to enforce
positivity. Note that this may spoil the convergence.
Must return `(x, fx, converged, tol)` where
- `x` is the root,
- `fx` is the residual at the root,
- `converged` indicates if the search was succesful.
- `tol` is the tolerance used for termination,
"""
abstract type Rootfinder end


find_root(f, x0_or_xrange, method) = error("'find_root' not implemented for \
                                           method '$(typeof(method))'")


#######################################################################
#                           Implementations                           #
#######################################################################


isfinite(x::Number) = !(isnan(x) || isinf(x))
tolerance(f, rf::Rootfinder) = rf.reltol * abs(f) + rf.abstol


#######################################################################
#                              Bisection                              #
#######################################################################


Base.@kwdef struct Bisection <: Rootfinder
  reltol::Float64 = 1e-10
  abstol::Float64 = 1e-10
  maxits::Int64   = 20
end
const Default_Bisection = Bisection()


"""
    find_root(f::Function, xrange, method::Bisection=Default_Bisection; sanitize=identity)

Find root of function `f` within `xrange`. `sanitize` is unused.

See also: `Rootfinder`.
"""
function find_root(f::Function, xrange, method::Bisection=Default_Bisection; 
    sanitize=identity #=unused=#)

  xl, xr = xrange
  if xr < xl
    error("bisection: Invalid search interval: ($xl, $xr)")
  end

  @unpack maxits = method

  fl, fr = first(f(xl)), first(f(xr))
  sign = fl < fr ? 1.0 : -1.0

  # setup
  x = (xl + xr) / 2
  fx = sign * first(f(x))
  tol = tolerance(fx, method)
  if abs(fx) <= tol
    return x, fx, true, tol
  end

  it = 1
  converged = false
  while !converged && it < maxits
    if fx < 0
      xl = x
    elseif fx > 0
      xr = x
    else
      converged = true
      break
    end
    x = (xl + xr) / 2
    fx = sign * first(f(x))
    it += 1
    if abs(fx) <= tol
      converged = true
    end
  end

  return x, fx, converged, tol
end


#######################################################################
#                            NewtonSolver                             #
#######################################################################


@with_kw struct NewtonSolver <: Rootfinder
  abstol::Float64 = 1e-10
  reltol::Float64 = 1e-10
  step::Float64   = 1e-8
  maxits::Int64   = 10
  approx_derivative::Bool = true
end
const Default_NewtonSolver = NewtonSolver()


"""
    find_root(f::Function, x0::Number, method::NewtonSolver=Default_NewtonSolver; sanitize=identity)

Apply Newton's method to find root of function `f` with starting guess `x0`.
Use `NewtonSolver(approx_derivative=true)` to utilize the analytic derivative, if provided. 

See also: `Rootfinder`.
"""
function find_root(f_df::Function, x0::Number, method::NewtonSolver=Default_NewtonSolver;
                 sanitize::Function=identity)
  if method.approx_derivative
    return _newton_solver_approx_derivative(f_df, x0, method; sanitize=sanitize)
  else
    return _newton_solver_exact_derivative(f_df, x0, method; sanitize=sanitize)
  end
end


function _newton_solver_approx_derivative(f_df::Function, x0,
    method::NewtonSolver=Default_NewtonSolver; sanitize::Function=identity)

  @unpack step, maxits = method

  xi = sanitize(x0)
  fi = first(f_df(xi))
  tol = tolerance(fi, method)

  if any(.!isfinite.((xi, fi)))
    return xi, fi, false, tol
  end
  if abs(fi) < tol
    return xi, fi, true, tol
  end

  it = 1
  converged = false
  while !converged && it <= maxits

    # scaling FD perturbation
    # C.T.Kelly - Solving Nonlinear Equations with Newton's Method, 2003
    # Section 2.3
    xdx, dx = if isapprox(x0, 0.0)
      dx = step
      dx, dx
    else
      dx = step * max(abs(x0), 1.0) * sign(x0)
      x0+dx, dx
    end
    dfi = (first(f_df(xdx)) - fi) / dx

    # update iterate
    xi = sanitize(x0 - fi / dfi)
    fi = first(f_df(xi))

    if any(.!isfinite.((xi, fi)))
      break
    end

    # accept step
    x0 = xi
    it += 1

    # test criterion
    if abs(fi) <= tol
      converged = true
    end
  end
  return xi, fi, converged, tol
end


function _newton_solver_exact_derivative(f_df::Function, x0,
    method::NewtonSolver=Default_NewtonSolver; sanitize::Function=identity)

  @unpack maxits = method

  xi = sanitize(x0)
  fi, dfi = f_df(xi)
  tol = tolerance(fi, method)
  if any(.!isfinite.((xi, fi, dfi)))
    return xi, fi, false
  end
  if abs(fi) <= tol
    return xi, fi, true, tol
  end

  it = 1
  converged = false
  while !converged && it <= maxits

    # update iterate
    xi = sanitize(x0 - fi / dfi)

    # compute residual
    fi, dfi = f_df(xi)

    # sanity check
    if any(.!isfinite.((xi,fi,dfi)))
      break
    end

    # accept step
    x0 = xi
    it += 1

    # test criterion
    if abs(fi) <= tol
      converged = true
    end
  end

  return xi, fi, converged, tol
end


#######################################################################
#                             Deprecated                              #
#######################################################################


@with_kw struct NlSolveParameters
  abstol::Float64 = 1e-10
  reltol::Float64 = 1e-10
  step::Float64   = 1e-8
  maxits::Int64   = 10
end
const Default_NlSolveParameters = NlSolveParameters()


function nlsolve(f_df::Function, x0; prms::NlSolveParameters=Default_NlSolveParameters, 
    sanitize=identity)

  newtonprms = NewtonSolver(prms.abstol, prms.reltol, prms.step, prms.maxits, true)
  find_root(f_df, x0, newtonprms; sanitize=sanitize)
end


function nlsolve_exact(f_df::Function, x0;
    prms::NlSolveParameters=Default_NlSolveParameters, sanitize=identity)

  newtonprms = NewtonSolver(prms.abstol, prms.reltol, prms.step, prms.maxits, false)
  find_root(f_df, x0, newtonprms; sanitize=sanitize)
end
