@testset "Evolution" begin


@testset "Convergence rate Runge Kutta schemes" begin

  # Test problem: df/dt = - α u = rhs(u), u(t0) = u0
  t0 = 0.0
  tend = 1.0
  u0 = [0.5]
  α = 0.8
  function rhs!(du, u, t)
    du .= - α * u
  end
  u(t) = u0 * exp(- α * (t - t0))

  u_ana = u(tend)

  schemes = [ :MIDPT, :RK4,       :RK4_twothirds, :RK4_Ralston, :SSPRK3, :LSERK4, :RKF10 ]
  orders  = [ 2,      4,          4,              4,            3,       4,       10     ]

  tspan = (t0, tend)
  dts = [0.1, 0.05, 0.025]

  err = zeros(length(dts), length(schemes))
  for (it, dt) in enumerate(dts)
    for (is, scheme) in enumerate(schemes)
      u_num = dg1d.evolve_ERK(rhs!, u0, dt, tspan, scheme)
      err[it,is] = norm(u_num - u_ana)
    end
  end

  for (is, scheme) in enumerate(schemes)
    for (it, dt) in enumerate(dts)
      if it < length(dts)
        e1 = err[it,is]
        e2 = err[end,is]
        if abs(e1 - e2) <= 1e-14 # cannot use such small errors to estimate convergence rate
          @test true
          continue
        end
        dt1 = dts[it]
        dt2 = dts[end]
        cvgrates = - log(e1/e2) / log(dt2/dt1)
        @test cvgrates > orders[is]
      end
    end
  end

end


@testset "Adaptive time stepping Runge Kutta scheme" begin

  # Test problem: df/dt = - α u = rhs(u), u(t0) = u0
  t0 = 0.0
  tend = 1.0
  u0 = [0.5]
  α = 15.0 # stiff ODE
  function rhs!(du, u, t)
    du .= - α * u
  end
  u(t) = u0 * exp(- α * (t - t0))

  u_ana = u(tend)

  schemes = [ :RK4, :RK4_twothirds, :RK4_Ralston, :SSPRK3, :LSERK4 ]
  tspan = (t0, tend)
  dts = [0.1, 0.05, 0.01]

  # solutions for different dts should all converge to the same result
  # hence, set reltol to zero to avoid influence from initial data
  reltol = 0.0
  abstol = 1e-14

  err = zeros(length(dts), length(schemes))
  for (it, dt) in enumerate(dts)
    for (is, scheme) in enumerate(schemes)
      u_num = dg1d.evolve_adaptive_ERK(rhs!, u0, dt, tspan, scheme=scheme,
                                       reltol=reltol, abstol=abstol)
      err[it,is] = norm(u_num - u_ana)
    end
  end

  # for (it, dt) in enumerate(dts)
  #   for (is, scheme) in enumerate(schemes)
  #     u_num = dg1d.evolve_adaptive_ERK(rhs!, u0, dt, tspan, scheme=scheme;
  #                                      reltol=reltol, abstol=abstol)
  #     err[it,is] = norm(u_num - u_ana, 1)
  #   end
  # end
  #
  # for (is, scheme) in enumerate(schemes)
  #   for (it, dt) in enumerate(dts)
  #     if it <= length(dts)
  #       # for now only verify if correct result is obtained for different
  #       # initial time steps
  #       @test err[it,is] <= 1e-10
  #     end
  #   end
  # end

end


end
